import { __rest } from 'tslib';
import { createContext, createElement, useContext, Fragment, useState, useEffect, forwardRef, useCallback, Component, useMemo, useRef, Suspense, memo } from 'react';
import loadable from '@loadable/component';
import { unionBy, get, pipe, omit, flattenDepth, throttle, last } from 'lodash/fp';
import { Link as Link$1, createHistory, LocationProvider, Router } from '@reach/router';
import merge from 'deepmerge';
import equal from 'fast-deep-equal';
import sort from 'array-sort';
import { ulid } from 'ulid';
import match from 'match-sorter';
import capitalize from 'capitalize';
import { MDXProvider } from '@mdx-js/tag';

const DefaultNotFound = () => createElement(Fragment, null, "Not found");
const DefaultLoading = () => createElement(Fragment, null, "Loading");
const DefaultPage = ({ children }) => createElement(Fragment, null, children);
const DefaultPlayground = ({ component, code }) => (createElement(Fragment, null,
    component,
    code));
const defaultComponents = {
    loading: DefaultLoading,
    playground: DefaultPlayground,
    notFound: DefaultNotFound,
    page: DefaultPage,
};
const ctx = createContext({});
const ComponentsProvider = ({ components: themeComponents = {}, children, }) => (createElement(ctx.Provider, { value: Object.assign({}, defaultComponents, themeComponents) }, children));
const useComponents = () => {
    return useContext(ctx);
};

const isFn = (value) => typeof value === 'function';
function flatArrFromObject(arr, prop) {
    const reducer = (arr, obj) => {
        const value = get(prop)(obj);
        return value ? arr.concat([value]) : arr;
    };
    return Array.from(new Set(arr.reduce(reducer, [])));
}
function compare(a, b, reverse) {
    if (a < b)
        return reverse ? 1 : -1;
    if (a > b)
        return reverse ? -1 : 1;
    return 0;
}
function mergeArrBy(prop, a, b) {
    return unionBy(prop, a)(b);
}

const AsyncComponent = defaultProps => {
    const [loading, setLoading] = useState(true);
    const [error, setError] = useState(null);
    const [data, setData] = useState({});
    useEffect(() => {
        const { getInitialProps } = defaultProps;
        if (getInitialProps && isFn(getInitialProps)) {
            setLoading(true);
            getInitialProps(defaultProps)
                .then((data) => {
                setLoading(false);
                setError(null);
                setData(data);
            })
                .catch((err) => {
                setLoading(false);
                setError(err);
                setData({});
            });
        }
    }, []);
    const { as: Comp, getInitialProps } = defaultProps, props = __rest(defaultProps, ["as", "getInitialProps"]);
    return createElement(Comp, Object.assign({}, props, { data: Object.assign({}, data, { loading, error }) }));
};

const loadRoute = (path, imports) => {
    return loadable(async () => {
        const { default: Component, getInitialProps } = await imports[path]();
        const ExportedComponent = props => (createElement(AsyncComponent, Object.assign({}, props, { as: Component || 'div', getInitialProps: getInitialProps })));
        return ExportedComponent;
    });
};
const AsyncRoute = defaultProps => {
    const { asyncComponent, path, entry } = defaultProps, routeProps = __rest(defaultProps, ["asyncComponent", "path", "entry"]);
    const components = useComponents();
    const Page = components.page;
    const Component = asyncComponent;
    const props = Object.assign({}, routeProps, { doc: entry });
    return Page ? (createElement(Page, Object.assign({}, props),
        createElement(Component, Object.assign({}, props)))) : (createElement(Component, Object.assign({}, props)));
};

const Link = forwardRef((props, ref) => {
    const isActive = useCallback(({ isCurrent }) => {
        return isCurrent ? { className: `${props.className} active` } : {};
    }, [props.className]);
    return createElement(Link$1, Object.assign({}, props, { getProps: isActive, ref: ref }));
});
Link.displayName = 'Link';

function create(initial) {
    var _a;
    const ctx = createContext(initial);
    const listeners = new Set();
    const dispatch = (fn) => {
        listeners.forEach(listener => listener(fn));
    };
    return {
        context: ctx,
        set: fn => dispatch(fn),
        Provider: (_a = class Provider extends Component {
                constructor() {
                    super(...arguments);
                    this.state = this.props.initial || initial || {};
                }
                static getDerivedStateFromProps(props, state) {
                    if (!equal(props.initial, state))
                        return props.initial;
                    return null;
                }
                componentDidMount() {
                    listeners.add((fn) => this.setState(fn));
                }
                componentWillUnmount() {
                    listeners.clear();
                }
                render() {
                    return (createElement(ctx.Provider, { value: this.state }, this.props.children));
                }
            },
            _a.displayName = 'DoczStateProvider',
            _a),
    };
}

// tslint:disable-next-line
const doczState = create({});

const useConfig = () => {
    const state = useContext(doczState.context);
    const { linkComponent, transform, config, themeConfig = {} } = state;
    const newConfig = merge(themeConfig, config ? config.themeConfig : {});
    const transformed = transform ? transform(newConfig) : newConfig;
    return Object.assign({}, config, { linkComponent, themeConfig: transformed });
};

const updateState = (ev) => {
    const { type, payload } = JSON.parse(ev.data);
    const prop = type.startsWith('state.') && type.split('.')[1];
    if (prop) {
        doczState.set(state => (Object.assign({}, state, { [prop]: payload })));
    }
};
const useDataServer = (url) => {
    useEffect(() => {
        if (!url)
            return;
        const socket = new WebSocket(url);
        socket.onmessage = updateState;
        return () => socket.close();
    }, []);
};

const useDocs = () => {
    const { entries, config } = useContext(doczState.context);
    if (!entries || !config)
        return null;
    const arr = entries.map(({ value }) => value);
    return sort(arr, (a, b) => compare(a.name, b.name));
};

const noMenu = (entry) => !entry.menu;
const fromMenu = (menu) => (entry) => entry.menu === menu;
const entryAsMenu = (entry) => ({
    name: entry.name,
    route: entry.route,
});
const entriesOfMenu = (menu, entries) => entries.filter(fromMenu(menu)).map(entryAsMenu);
const parseMenu = (entries) => (name) => ({
    name,
    menu: entriesOfMenu(name, entries),
});
const menusFromEntries = (entries) => {
    const entriesWithoutMenu = entries.filter(noMenu).map(entryAsMenu);
    const menus = flatArrFromObject(entries, 'menu').map(parseMenu(entries));
    return [...entriesWithoutMenu, ...menus];
};
const parseItemStr = (item) => typeof item === 'string' ? { name: item } : item;
const normalize = (item) => {
    const selected = parseItemStr(item);
    return Object.assign({}, selected, { id: selected.id || ulid(), menu: Array.isArray(selected.menu)
            ? selected.menu.map(normalize)
            : selected.menu });
};
const clean = (item) => item.href || item.route ? omit('menu', item) : item;
const normalizeAndClean = pipe(normalize, clean);
const mergeMenus = (entriesMenu, configMenu) => {
    const first = entriesMenu.map(normalizeAndClean);
    const second = configMenu.map(normalizeAndClean);
    const merged = mergeArrBy('name', first, second);
    return merged.map(item => {
        if (!item.menu)
            return item;
        const found = second.find(i => i.name === item.name);
        const foundMenu = found && found.menu;
        return Object.assign({}, item, { menu: foundMenu
                ? mergeMenus(item.menu, foundMenu)
                : item.menu || found.menu });
    });
};
const UNKNOWN_POS = Infinity;
const findPos = (item, orderedList = []) => {
    const name = typeof item !== 'string' ? get('name', item) : item;
    const pos = orderedList.findIndex(item => item === name);
    return pos !== -1 ? pos : UNKNOWN_POS;
};
const compareWithMenu = (to = []) => (a, b) => {
    const list = to.map((i) => i.name || i);
    return compare(findPos(a, list), findPos(b, list));
};
const sortByName = (a, b) => {
    return a.name < b.name ? -1 : a.name > b.name ? 1 : 0;
};
const sortMenus = (first, second = []) => {
    const sorted = sort(first, compareWithMenu(second), sortByName);
    return sorted.map(item => {
        if (!item.menu)
            return item;
        const found = second.find(menu => menu.name === item.name);
        const foundMenu = found && found.menu;
        return Object.assign({}, item, { menu: foundMenu
                ? sortMenus(item.menu, foundMenu)
                : sort(item.menu, sortByName) });
    });
};
const search = (val, menu) => {
    const items = menu.map(item => [item].concat(item.menu || []));
    const flattened = flattenDepth(2, items);
    const flattenedDeduplicated = [...new Set(flattened)];
    return match(flattenedDeduplicated, val, { keys: ['name'] });
};
const useMenus = ({ query = '' }) => {
    const { entries, config } = useContext(doczState.context);
    if (!entries || !config)
        return null;
    const arr = entries.map(({ value }) => value);
    const entriesMenu = menusFromEntries(arr);
    const sorted = useMemo(() => {
        const merged = mergeMenus(entriesMenu, config.menu);
        return sortMenus(merged, config.menu);
    }, [entries, config]);
    return query.length > 0 ? search(query, sorted) : sorted;
};

const usePrevious = (value, defaultValue) => {
    const ref = useRef(defaultValue);
    useEffect(() => {
        ref.current = value;
    });
    return ref.current;
};

const getSize = () => ({
    innerHeight: window.innerHeight,
    innerWidth: window.innerWidth,
    outerHeight: window.outerHeight,
    outerWidth: window.outerWidth,
});
const useWindowSize = () => {
    const [windowSize, setWindowSize] = useState(getSize());
    const tSetWindowResize = throttle(300, () => setWindowSize(getSize()));
    useEffect(() => {
        window.addEventListener('resize', tSetWindowResize);
        return () => void window.removeEventListener('resize', tSetWindowResize);
    }, []);
    return windowSize;
};

const Playground = ({ className, style, wrapper: Wrapper, children, __scope, __position, __code, __codesandbox, }) => {
    const components = useComponents();
    if (!components || !components.playground)
        return null;
    const props = { className, style, components };
    return (createElement(components.playground, Object.assign({}, props, { component: children, wrapper: Wrapper, scope: __scope, position: __position, code: __code, codesandbox: __codesandbox })));
};

const RE_OBJECTOF = /(?:React\.)?(?:PropTypes\.)?objectOf\((?:React\.)?(?:PropTypes\.)?(\w+)\)/;
const getTypeStr = (type) => {
    switch (type.name.toLowerCase()) {
        case 'instanceof':
            return `Class(${type.value})`;
        case 'enum':
            if (type.computed)
                return type.value;
            return type.value
                ? type.value.map((v) => `${v.value}`).join(' │ ')
                : type.raw;
        case 'union':
            return type.value
                ? type.value.map((t) => `${getTypeStr(t)}`).join(' │ ')
                : type.raw;
        case 'array':
            return type.raw;
        case 'arrayof':
            return `Array<${getTypeStr(type.value)}>`;
        case 'custom':
            if (type.raw.indexOf('function') !== -1 || type.raw.indexOf('=>') !== -1)
                return 'Custom(Function)';
            else if (type.raw.toLowerCase().indexOf('objectof') !== -1) {
                const m = type.raw.match(RE_OBJECTOF);
                if (m && m[1])
                    return `ObjectOf(${capitalize(m[1])})`;
                return 'ObjectOf';
            }
            return 'Custom';
        case 'bool':
            return 'Boolean';
        case 'func':
            return 'Function';
        case 'shape':
            const shape = type.value;
            const rst = {};
            Object.keys(shape).forEach(key => {
                rst[key] = getTypeStr(shape[key]);
            });
            return JSON.stringify(rst, null, 2);
        default:
            return capitalize(type.name);
    }
};
const humanize = (type) => getTypeStr(type);

const getPropType = (prop) => {
    const propName = get('name', prop.flowType || prop.type);
    if (!propName)
        return null;
    const isEnum = propName.startsWith('"') || propName === 'enum';
    const name = capitalize(isEnum ? 'enum' : propName);
    const value = get('type.value', prop);
    if (!name)
        return null;
    if ((isEnum && typeof value === 'string') ||
        (!prop.flowType && !isEnum && !value) ||
        (prop.flowType && !prop.flowType.elements)) {
        return name;
    }
    return prop.flowType ? humanize(prop.flowType) : humanize(prop.type);
};
const Props = ({ of: component }) => {
    const components = useComponents();
    const { props: stateProps } = useContext(doczState.context);
    const filename = get('__filemeta.filename', component);
    const found = stateProps &&
        stateProps.length > 0 &&
        stateProps.find(item => item.key === filename);
    const definition = last(found ? found.value : []);
    const props = get('props', definition) || [];
    if (!props)
        return null;
    if (!components.props)
        return null;
    return createElement(components.props, { props: props, getPropType: getPropType });
};

const goToHash = ({ location }) => {
    setTimeout(() => {
        if (location && location.hash) {
            const id = location.hash.substring(1);
            const el = document.getElementById(id);
            if (el)
                el.scrollIntoView();
        }
    });
};
const Routes = ({ imports }) => {
    const components = useComponents();
    const { entries } = useContext(doczState.context);
    if (!entries || !components)
        return null;
    const NotFound = components.notFound;
    const Loading = components.loading;
    const history = useMemo(() => createHistory(window), []);
    useEffect(() => {
        history.listen(goToHash);
    }, []);
    return (createElement(MDXProvider, { components: components },
        createElement(LocationProvider, { history: history },
            createElement(Suspense, { fallback: createElement(Loading, null) },
                createElement(Router, { basepath: DOCZ_BASE_URL },
                    createElement(NotFound, { default: true }),
                    entries.map(({ key: path, value: entry }) => {
                        const props = { path, entries, components };
                        const component = loadRoute(path, imports);
                        return (createElement(AsyncRoute, Object.assign({}, props, { entry: entry, key: entry.id, path: entry.route, asyncComponent: component })));
                    }))))));
};

function theme(themeConfig, transform = c => c) {
    return WrappedComponent => {
        const Theme = memo(props => {
            const { linkComponent } = props;
            const { db, children, wrapper: Wrapper = Fragment } = props;
            const initial = Object.assign({}, db, { themeConfig, transform, linkComponent });
            return (createElement(doczState.Provider, { initial: initial },
                createElement(Wrapper, null,
                    createElement(WrappedComponent, null, children))));
        });
        return Theme;
    };
}

export { AsyncRoute, loadRoute, Link, Playground, Props, Routes, theme, doczState, useConfig, useDataServer, useDocs, useMenus, usePrevious, useWindowSize, ComponentsProvider, useComponents };
